<?php

namespace Drupal\commerce_adyen_cc;

use Adyen\AdyenException;
use Drupal\commerce_adyen_cc\Enums\ResultCode;
use Drupal\commerce_payment\Exception\InvalidRequestException;
use Drupal\commerce_payment\Exception\InvalidResponseException;
use Drupal\commerce_payment\Exception\SoftDeclineException;

/**
 * Translates Adyen exceptions and errors into Commerce exceptions.
 *
 * @see https://docs.adyen.com/online-payments/payment-result-codes
 * @see https://docs.adyen.com/development-resources/error-codes
 */
class ErrorHelper {

  /**
   * Translates Adyen exceptions into Commerce exceptions.
   *
   * @param \Adyen\AdyenException $exception
   *   The Adyen exception.
   *
   * @throws \Drupal\commerce_payment\Exception\PaymentGatewayException
   *   The Commerce exception.
   */
  public static function handleException(AdyenException $exception): void {
    throw new InvalidResponseException($exception->getMessage());
  }

  /**
   * Translates Adyen errors into Commerce exceptions.
   *
   * @param array $result
   *   The Adyen result.
   *
   * @throws \Drupal\commerce_payment\Exception\PaymentGatewayException
   *   The Commerce exception.
   */
  public static function handleErrors(array $result): void {
    $result_code = $result['resultCode'] ?? $result['status'] ?? 'unknown';
    if (in_array($result_code, [
      ResultCode::AUTHORIZED,
      ResultCode::AUTHENTICATION_FINISHED,
      ResultCode::RECEIVED,
    ], TRUE)) {
      return;
    }
    switch ($result['resultCode']) {
      case ResultCode::REFUSED:
        $message = $result['refusalReason'] ?? 'Unknown';
        $refusal_reason_code = $result['refusalReasonCode'] ?? '';
        if (!empty($refusal_reason_code)) {
          $message .= ' (' . $refusal_reason_code . ')';
        }
        throw new SoftDeclineException(trim($message));
    }

    // Throw a fallback exception for everything else.
    throw new InvalidRequestException('Invalid Result Code: ' . $result_code);
  }

}
